from pypanther.rules.azure_signin.azure_failed_signins import AzureAuditManyFailedSignIns as AzureAuditManyFailedSignIns
from pypanther.rules.azure_signin.azure_legacyauth import AzureAuditLegacyAuth as AzureAuditLegacyAuth
from pypanther.rules.azure_signin.azure_risklevel_passthrough import (
    AzureAuditRiskLevelPassthrough as AzureAuditRiskLevelPassthrough,
)
